"""
General uncategorised functions/classes for network related helper code

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""
import asyncio
import logging
import random
import socket
from datetime import datetime
from math import ceil
from typing import List, Tuple

from privex.helpers.decorators import r_cache, r_cache_async

from privex.helpers import settings
from privex.helpers.common import byteify, empty, empty_if, is_true
from privex.helpers.asyncx import run_coro_thread_async
from privex.helpers.net import base as netbase
from privex.helpers.net.dns import resolve_ip, resolve_ip_async
from privex.helpers.net.socket import AsyncSocketWrapper
from privex.helpers.net.util import get_ssl_context, ip_is_v6
from privex.helpers.types import AUTO, AnyNum, IP_OR_STR

log = logging.getLogger(__name__)

__all__ = [
    'check_host', 'check_host_async', 'check_host_http', 'check_host_http_async', 'test_hosts_async',
    'test_hosts', 'check_v4', 'check_v6', 'check_v4_async', 'check_v6_async'
]


def check_host(host: IP_OR_STR, port: AnyNum, version='any', throw=False, **kwargs) -> bool:
    """
    Test if the service on port ``port`` for host ``host`` is working. AsyncIO version: :func:`.check_host_async`
    
    Basic usage (services which send the client data immediately after connecting)::
    
        >>> check_host('hiveseed-se.privex.io', 2001)
        True
        >>> check_host('hiveseed-se.privex.io', 9991)
        False
    
    For some services, such as HTTP - it's necessary to transmit some data to the host before it will
    send a response. Using the ``send`` kwarg, you can transmit an arbitrary string/bytes upon connection.
    
    Sending data to ``host`` after connecting::
    
        >>> check_host('files.privex.io', 80, send=b"GET / HTTP/1.1\\n\\n")
        True
    
    
    :param str|IPv4Address|IPv6Address host: Hostname or IP to test
    :param int|str port: Port number on ``host`` to connect to
    :param str|int version: When connecting to a hostname, this can be set to ``'v4'``, ``'v6'`` or similar
                            to ensure the connection is via that IP version
    
    :param bool throw: (default: ``False``) When ``True``, will raise exceptions instead of returning ``False``
    :param kwargs: Additional configuration options (see below)
    
    :keyword int receive: (default: ``100``) Amount of bytes to attempt to receive from the server (``0`` to disable)
    :keyword bytes|str send: If ``send`` is specified, the data in ``send`` will be transmitted to the server before receiving.
    :keyword int stype: Socket type, e.g. :attr:`socket.SOCK_STREAM`
    
    :keyword float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                                If the global default timeout is ``None``, then falls back to ``5.0``
    
    :raises socket.timeout: When ``throw=True`` and a timeout occurs.
    :raises socket.gaierror: When ``throw=True`` and various errors occur
    :raises ConnectionRefusedError: When ``throw=True`` and the connection was refused
    :raises ConnectionResetError: When ``throw=True`` and the connection was reset
    
    :return bool success: ``True`` if successfully connected + sent/received data. Otherwise ``False``.
    """
    receive, stype = int(kwargs.get('receive', 100)), kwargs.get('stype', socket.SOCK_STREAM)
    timeout, send, use_ssl = kwargs.get('timeout', 'n/a'), kwargs.get('send'), kwargs.get('ssl', kwargs.get('use_ssl'))
    ssl_params = kwargs.get('ssl_params', dict(verify_cert=False, check_hostname=False))
    if timeout == 'n/a':
        t = socket.getdefaulttimeout()
        timeout = 10.0 if not t else t
    
    try:
        s_ver = socket.AF_INET
        ip = resolve_ip(host, version)
    
        if ip_is_v6(ip): s_ver = socket.AF_INET6
    
        if port == 443 and use_ssl is None:
            log.warning("check_host: automatically setting use_ssl=True as port is 443 and use_ssl was not specified.")
            use_ssl = True
        with socket.socket(s_ver, stype) as s:
            orig_sock = s
            if timeout: s.settimeout(float(timeout))
            if use_ssl:
                ctx = get_ssl_context(**ssl_params)
                s = ctx.wrap_socket(
                    s,
                    server_hostname=kwargs.get('server_hostname'),
                    session=kwargs.get('session'),
                    do_handshake_on_connect=kwargs.get('do_handshake_on_connect', True),
                )
                
            s.connect((ip, int(port)))
            if not empty(send):
                s.sendall(byteify(send))
            if receive > 0:
                s.recv(int(receive))
            if use_ssl:
                s.close()
        return True
    except (socket.timeout, TimeoutError, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        if throw:
            raise e
    return False


async def check_host_async(host: IP_OR_STR, port: AnyNum, version='any', throw=False, **kwargs) -> bool:
    """
    AsyncIO version of :func:`.check_host`. Test if the service on port ``port`` for host ``host`` is working.

    Basic usage (services which send the client data immediately after connecting)::

        >>> await check_host_async('hiveseed-se.privex.io', 2001)
        True
        >>> await check_host_async('hiveseed-se.privex.io', 9991)
        False

    For some services, such as HTTP - it's necessary to transmit some data to the host before it will
    send a response. Using the ``send`` kwarg, you can transmit an arbitrary string/bytes upon connection.

    Sending data to ``host`` after connecting::

        >>> await check_host_async('files.privex.io', 80, send=b"GET / HTTP/1.1\\n\\n")
        True


    :param str|IPv4Address|IPv6Address host: Hostname or IP to test
    :param int|str port: Port number on ``host`` to connect to
    :param str|int version: When connecting to a hostname, this can be set to ``'v4'``, ``'v6'`` or similar
                            to ensure the connection is via that IP version

    :param bool throw: (default: ``False``) When ``True``, will raise exceptions instead of returning ``False``
    :param kwargs: Additional configuration options (see below)

    :keyword int receive: (default: ``100``) Amount of bytes to attempt to receive from the server (``0`` to disable)
    :keyword bytes|str send: If ``send`` is specified, the data in ``send`` will be transmitted to the server before receiving.
    :keyword int stype: Socket type, e.g. :attr:`socket.SOCK_STREAM`

    :keyword float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                                If the global default timeout is ``None``, then falls back to ``5.0``

    :raises socket.timeout: When ``throw=True`` and a timeout occurs.
    :raises socket.gaierror: When ``throw=True`` and various errors occur
    :raises ConnectionRefusedError: When ``throw=True`` and the connection was refused
    :raises ConnectionResetError: When ``throw=True`` and the connection was reset

    :return bool success: ``True`` if successfully connected + sent/received data. Otherwise ``False``.
    """
    receive, stype = int(kwargs.get('receive', 16)), kwargs.get('stype', socket.SOCK_STREAM)
    timeout, send = kwargs.get('timeout', 'n/a'), kwargs.get('send')
    http_test, use_ssl = kwargs.get('http_test', False), kwargs.get('use_ssl', False)
    if timeout == 'n/a':
        t = socket.getdefaulttimeout()
        timeout = settings.DEFAULT_SOCKET_TIMEOUT if not t else t
    
    # loop = asyncio.get_event_loop()
    s_ver = socket.AF_INET
    ip = await resolve_ip_async(host, version)
    
    if ip_is_v6(ip): s_ver = socket.AF_INET6
    
    try:
        aw = AsyncSocketWrapper(host, int(port), family=s_ver, use_ssl=use_ssl, timeout=timeout)
        await aw.connect()
        if http_test:
            log.info("Sending HTTP request to %s", host)
            log.info("Response from %s : %s", host, await aw.http_request())
        
        elif not empty(send) and receive > 0:
            log.info("Sending query data '%s' and trying to receive data from %s", send, host)
            log.info("Response from %s : %s", host, await aw.query(send, receive, read_timeout=kwargs.get('read_timeout', AUTO)))
        
        elif not empty(send):
            log.info("Sending query data '%s' to %s", send, host)
            await aw.sendall(send)
        else:
            log.info("Receiving data from %s", host)
            
            log.info("Response from %s : %s", host, await aw.read_eof(
                receive, strip=False, read_timeout=kwargs.get('read_timeout', AUTO),
            ))
        
        # with socket.socket(s_ver, stype) as s:
        #     if timeout: s.settimeout(float(timeout))
        #     await loop.sock_connect(s, (ip, int(port)))
        #     if not empty(send):
        #         await loop.sock_sendall(s, byteify(send))
        #     if receive > 0:
        #         await loop.sock_recv(s, int(receive))
        return True
    except (socket.timeout, TimeoutError, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        if throw:
            raise e
    return False


def check_host_http(host: IP_OR_STR, port: AnyNum = 80, version='any', throw=False, **kwargs) -> bool:
    return netbase.check_host(host, port, version, throw=throw, http_test=True, **kwargs)


async def check_host_http_async(
            host: IP_OR_STR, port: AnyNum = 80, version='any', throw=False, send=b"GET / HTTP/1.1\\n\\n", **kwargs
        ) -> bool:
    # return await check_host_async(host, port, version, throw=throw, send=send, **kwargs)
    return await netbase.check_host_async(host, port, version, throw=throw, http_test=True, **kwargs)


async def test_hosts_async(hosts: List[str] = None, ipver: str = 'any', timeout: AnyNum = None, **kwargs) -> bool:
    randomise = is_true(kwargs.get('randomise', True))
    max_hosts = kwargs.get('max_hosts', settings.NET_CHECK_HOST_COUNT_TRY)
    if max_hosts is not None: max_hosts = int(max_hosts)
    timeout = empty_if(timeout, empty_if(socket.getdefaulttimeout(), 4, zero=True), zero=True)
    
    v4h, v6h = list(settings.V4_TEST_HOSTS), list(settings.V6_TEST_HOSTS)
    if randomise: random.shuffle(v4h)
    if randomise: random.shuffle(v6h)
    
    if empty(hosts, True, True):
        # if empty(ipver, True, True) or ipver in ['any', 'all', 'both', 10, '10', '46', 46]:
        #     settings.V4_CHECKED_AT
        if isinstance(ipver, str): ipver = ipver.lower()
        if ipver in [4, '4', 'v4', 'ipv4']:
            hosts = v4h
            ipver = 4
        elif ipver in [6, '6', 'v6', 'ipv6']:
            hosts = v6h
            ipver = 6
        else:
            ipver = 'any'
            if max_hosts:
                hosts = v4h[:int(ceil(max_hosts / 2))] + v6h[:int(ceil(max_hosts / 2))]
            else:
                hosts = v4h + v6h
    
    if max_hosts: hosts = hosts[:max_hosts]
    
    # st4_empty = any([empty(settings.HAS_WORKING_V4, True, True), empty(settings.V4_CHECKED_AT, True, True)])
    # st6_empty = any([empty(settings.HAS_WORKING_V6, True, True), empty(settings.V6_CHECKED_AT, True, True)])
    
    # if ipver == 6 and not st6_empty and settings.V6_CHECKED_AT > datetime.utcnow():
    #     # if settings.V6_CHECKED_AT > datetime.utcnow()
    #     log.debug("Returning cached IPv6 status: working = %s", settings.HAS_WORKING_V6)
    #     return settings.HAS_WORKING_V6
    # if ipver == 4 and not st4_empty and settings.V4_CHECKED_AT > datetime.utcnow():
    #     # if settings.V6_CHECKED_AT > datetime.utcnow()
    #     log.debug("Returning cached IPv4 status: working = %s", settings.HAS_WORKING_V4)
    #     return settings.HAS_WORKING_V4
    #
    # if ipver == 'any' and any([not st4_empty, not st6_empty]) and settings.V4_CHECKED_AT > datetime.utcnow():
    #     # if settings.V6_CHECKED_AT > datetime.utcnow()
    #     if st4_empty:
    #         log.debug("test_hosts being requested for 'any' ip ver. IPv6 status cached, but not IPv4 status. Checking IPv4 status...")
    #         await check_v4_async()
    #     if st6_empty:
    #         log.debug("test_hosts being requested for 'any' ip ver. IPv4 status cached, but not IPv6 status. Checking IPv6 status...")
    #         await check_v6_async(hosts)
    #     # if not st4_empty and not st6_empty:
    #     log.debug(
    #         "Returning status %s based on: Working IPv4 = %s || Working IPv6 = %s",
    #         settings.HAS_WORKING_V4 or settings.HAS_WORKING_V6, settings.HAS_WORKING_V4, settings.HAS_WORKING_V6
    #     )
    #     return settings.HAS_WORKING_V4 or settings.HAS_WORKING_V6
    
    # max_hosts = int(kwargs.get('max_hosts', settings.NET_CHECK_HOST_COUNT_TRY))
    min_hosts_pos = int(kwargs.get('required_positive', settings.NET_CHECK_HOST_COUNT))
    
    # hosts = empty_if(hosts, settings.V4_TEST_HOSTS, itr=True)
    hosts = [x for x in hosts]
    
    if randomise: random.shuffle(hosts)
    
    if len(hosts) > max_hosts: hosts = hosts[:max_hosts]
    
    # port = empty_if(port, 80, zero=True)
    
    total_hosts = len(hosts)
    total_working, total_broken = 0, 0
    working_list, broken_list = [], []
    log.debug("Testing %s hosts with IP version '%s' - timeout: %s", total_hosts, ipver, timeout)

    host_checks = []
    host_checks_hosts = []
    for h in hosts:
        # host_checks.append(
        #     asyncio.create_task(_test_host_async(h, ipver=ipver, timeout=timeout))
        # )
        host_checks.append(
            asyncio.create_task(
                run_coro_thread_async(_test_host_async, h, ipver=ipver, timeout=timeout)
            )
        )
        host_checks_hosts.append(h)

    host_checks_res = await asyncio.gather(*host_checks, return_exceptions=True)
    for i, _res in enumerate(host_checks_res):
        h = host_checks_hosts[i]
        if isinstance(_res, Exception):
            log.warning("Exception while checking host %s", h)
            total_broken += 1
            continue

        res, h, port = _res
    
        if res:
            total_working += 1
            working_list.append(f"{h}:{port}")
            log.debug("check_host for %s (port %s) came back True (WORKING). incremented working hosts: %s", h, port, total_working)
        else:
            total_broken += 1
            broken_list.append(f"{h}:{port}")
            log.debug("check_host for %s (port %s) came back False (! BROKEN !). incremented broken hosts: %s", h, port, total_broken)

    # port = 80
    # for h in hosts:
    #     try:
    #         h, port, res = await _test_host_async(h, ipver, timeout)
    #         if res:
    #             total_working += 1
    #             log.debug("check_host for %s came back true. incremented working hosts: %s", h, total_working)
    #         else:
    #             total_broken += 1
    #             log.debug("check_host for %s came back false. incremented broken hosts: %s", h, total_broken)
    #
    #     except Exception as e:
    #         log.warning("Exception while checking host %s port %s", h, port)
    
    working = total_working >= min_hosts_pos
    
    log.info("test_hosts - proto: %s - protocol working? %s || total hosts: %s || working hosts: %s || broken hosts: %s",
             ipver, working, total_hosts, total_working, total_broken)
    log.debug("working hosts: %s", working_list)
    log.debug("broken hosts: %s", broken_list)
    
    return working


async def _test_host_async(host, ipver: str = 'any', timeout: AnyNum = None) -> Tuple[bool, str, int]:
    nh = host.split(':')
    if len(nh) > 1:
        port = int(nh[-1])
        host = ':'.join(nh[:-1])
    else:
        host = ':'.join(nh)
        log.warning("Host is missing port: %s - falling back to port 80")
        port = 80
    log.debug("Checking host %s via port %s + IP version '%s'", host, port, ipver)
    if port == 80:
        res = await check_host_http_async(host, port, ipver, throw=False, timeout=timeout)
    elif port == 53:
        res = await netbase.check_host_async(host, port, ipver, throw=False, timeout=timeout, send="hello\nworld\n")
    else:
        res = await netbase.check_host_async(host, port, ipver, throw=False, timeout=timeout)
    return res, host, port


def test_hosts(hosts: List[str] = None, ipver: str = 'any', timeout: AnyNum = None, **kwargs) -> bool:
    randomise = is_true(kwargs.get('randomise', True))
    max_hosts = kwargs.get('max_hosts', settings.NET_CHECK_HOST_COUNT_TRY)
    if max_hosts is not None: max_hosts = int(max_hosts)
    timeout = empty_if(timeout, empty_if(socket.getdefaulttimeout(), 4, zero=True), zero=True)

    v4h, v6h = list(settings.V4_TEST_HOSTS), list(settings.V6_TEST_HOSTS)
    if randomise: random.shuffle(v4h)
    if randomise: random.shuffle(v6h)

    if empty(hosts, True, True):
        # if empty(ipver, True, True) or ipver in ['any', 'all', 'both', 10, '10', '46', 46]:
        #     settings.V4_CHECKED_AT
        if isinstance(ipver, str): ipver = ipver.lower()
        if ipver in [4, '4', 'v4', 'ipv4']:
            hosts = v4h
            ipver = 4
        elif ipver in [6, '6', 'v6', 'ipv6']:
            hosts = v6h
            ipver = 6
        else:
            ipver = 'any'
            if max_hosts:
                hosts = v4h[:int(ceil(max_hosts / 2))] + v6h[:int(ceil(max_hosts / 2))]
            else:
                hosts = v4h + v6h

    if max_hosts: hosts = hosts[:max_hosts]
    
    # st4_empty = any([empty(settings.HAS_WORKING_V4, True, True), empty(settings.V4_CHECKED_AT, True, True)])
    # st6_empty = any([empty(settings.HAS_WORKING_V6, True, True), empty(settings.V6_CHECKED_AT, True, True)])

    # if ipver == 6 and not st6_empty and settings.V6_CHECKED_AT > datetime.utcnow():
    #     # if settings.V6_CHECKED_AT > datetime.utcnow()
    #     log.debug("Returning cached IPv6 status: working = %s", settings.HAS_WORKING_V6)
    #     return settings.HAS_WORKING_V6
    # if ipver == 4 and not st4_empty and settings.V4_CHECKED_AT > datetime.utcnow():
    #     # if settings.V6_CHECKED_AT > datetime.utcnow()
    #     log.debug("Returning cached IPv4 status: working = %s", settings.HAS_WORKING_V4)
    #     return settings.HAS_WORKING_V4

    # if ipver == 'any' and any([not st4_empty, not st6_empty]) and settings.V4_CHECKED_AT > datetime.utcnow():
    #     # if settings.V6_CHECKED_AT > datetime.utcnow()
    #     if st4_empty:
    #         log.debug("test_hosts being requested for 'any' ip ver. IPv6 status cached, but not IPv4 status. Checking IPv4 status...")
    #         check_v4()
    #     if st6_empty:
    #         log.debug("test_hosts being requested for 'any' ip ver. IPv4 status cached, but not IPv6 status. Checking IPv6 status...")
    #         check_v6()
    #     # if not st4_empty and not st6_empty:
    #     log.debug(
    #         "Returning status %s based on: Working IPv4 = %s || Working IPv6 = %s",
    #         settings.HAS_WORKING_V4 or settings.HAS_WORKING_V6, settings.HAS_WORKING_V4, settings.HAS_WORKING_V6
    #     )
    #     return settings.HAS_WORKING_V4 or settings.HAS_WORKING_V6

    # max_hosts = int(kwargs.get('max_hosts', settings.NET_CHECK_HOST_COUNT_TRY))
    min_hosts_pos = int(kwargs.get('required_positive', settings.NET_CHECK_HOST_COUNT))
    
    # hosts = empty_if(hosts, settings.V4_TEST_HOSTS, itr=True)
    hosts = [x for x in hosts]
    
    if randomise: random.shuffle(hosts)
    
    if len(hosts) > max_hosts: hosts = hosts[:max_hosts]
    
    
    total_hosts = len(hosts)
    total_working, total_broken = 0, 0

    log.debug("Testing %s hosts with IP version '%s' - timeout: %s", total_hosts, ipver, timeout)
    port = 80
    
    for h in hosts:
        try:
            nh = h.split(':')
            if len(nh) > 1:
                port = int(nh[-1])
                h = ':'.join(nh[:-1])
            else:
                h = ':'.join(nh)
                log.warning("Host is missing port: %s - falling back to port 80")
                port = 80
    
            log.debug("Checking host %s via port %s + IP version '%s'", h, port, ipver)

            if port == 80:
                res = check_host_http(h, port, ipver, throw=False, timeout=timeout)
            else:
                res = check_host(h, port, ipver, throw=False, timeout=timeout)
            if res:
                total_working += 1
                log.debug("check_host for %s came back true. incremented working hosts: %s", h, total_working)
            else:
                total_broken += 1
                log.debug("check_host for %s came back false. incremented broken hosts: %s", h, total_broken)

        except Exception as e:
            log.warning("Exception while checking host %s port %s", h, port)
    
    working = total_working >= min_hosts_pos
    
    log.info("test_hosts - proto: %s - protocol working? %s || total hosts: %s || working hosts: %s || broken hosts: %s",
             ipver, working, total_hosts, total_working, total_broken)
    
    return working


@r_cache("pvxhelpers:check_v4", settings.NET_CHECK_TIMEOUT)
def check_v4(hosts: List[str] = None, *args, **kwargs) -> bool:
    """Check and cache whether IPv4 is functional by testing a handful of IPv4 hosts"""
    return test_hosts(hosts, ipver='v4', *args, **kwargs)


@r_cache("pvxhelpers:check_v6", settings.NET_CHECK_TIMEOUT)
def check_v6(hosts: List[str] = None, *args, **kwargs) -> bool:
    """Check and cache whether IPv6 is functional by testing a handful of IPv6 hosts"""
    return test_hosts(hosts, ipver='v6', *args, **kwargs)


@r_cache_async("pvxhelpers:check_v4", settings.NET_CHECK_TIMEOUT)
async def check_v4_async(hosts: List[str] = None, *args, **kwargs) -> bool:
    """(Async ver of :func:`.check_v4`) Check and cache whether IPv4 is functional by testing a handful of IPv4 hosts"""
    return await test_hosts_async(hosts, ipver='v4', *args, **kwargs)


@r_cache_async("pvxhelpers:check_v6", settings.NET_CHECK_TIMEOUT)
async def check_v6_async(hosts: List[str] = None, *args, **kwargs) -> bool:
    """(Async ver of :func:`.check_v6`) Check and cache whether IPv6 is functional by testing a handful of IPv6 hosts"""
    return await test_hosts_async(hosts, ipver='v6', *args, **kwargs)



"""
Configuration options for helpers, and services they depend on, such as Redis.

To override settings from your app::

    >>> from privex.helpers import settings
    >>> settings.REDIS_HOST = 'redis.example.org'
    >>> settings.REDIS_PORT = 1234


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import random
import sys
from datetime import datetime
from os import getcwd, getenv as env
from os import path
from os.path import basename, dirname, abspath, join, expanduser
from typing import Optional

BASE_DIR = dirname(dirname(dirname(abspath(__file__))))
"""The root folder of this project (i.e. where setup.py is)"""

VERSION_FILE = join(BASE_DIR, 'privex', 'helpers', '__init__.py')
"""The file containing the package version (for :py:func:`.setuppy.bump_version`)"""

EXTRAS_FOLDER = 'extras'
"""Folder where additional requirements files can be found for :py:func:`privex.helpers.setuppy.common.extras`"""


def _is_true(v):
    return (v.lower() if type(v) is str else v) in [True, 'true', 'yes', 'y', '1', 1]


def _env_bool(v, d) -> bool: return _is_true(env(v, d))


def _env_int(v, d) -> int: return int(env(v, d))


########################################
#                                      #
#       Cache Module Settings          #
#                                      #
#       privex.helpers.cache           #
#                                      #
########################################

########
# General Cache Settings
########

DEFAULT_CACHE_TIMEOUT = _env_int('PRIVEX_CACHE_TIMEOUT', 300)

"""Default cache timeout in seconds, used by cache adapters in the module :py:mod:`.cache`"""

DEFAULT_CACHE_ADAPTER = env('PRIVEX_CACHE_ADAPTER', 'memory')
"""
The default cache adapter to use by default, with the framework-agnostic cache abstraction layer ( :mod:`privex.helpers.cache` ).

Can be one of the following:

  ( more options may be available since the time of writing, see :attr:`privex.helpers.cache.ADAPTER_MAP` )
  
  * ``memory`` / ``mem`` / ``ram`` - Dependency-free memory cache inside of your application's memory.
  
    **Requires:** N/A - no dependencies
    
    **Class:** :class:`.MemoryCache` (sync) // :class:`.AsyncMemoryCache` (async)
    
  * ``redis`` - Caching via a local or remote Redis server
     
     **Requires**: ``redis`` (sync) and/or ``aioredis`` (async) libraries
     
     **Class:** :class:`.RedisCache` (sync) // :class:`.AsyncRedisCache` (async)
     
  * ``memcached`` / ``memcache`` / ``mcache`` - Caching via a local or remote Memcached server
  
    **Requires**: ``pylibmc`` (sync) and/or // ``aiomcache`` (async) libraries
    
    **Class:** :class:`.MemcachedCache` (sync) // :class:`.AsyncMemcachedCache` (async)

  * ``sqlite3`` / ``sqlite`` / ``sqlitedb`` - Caching via an SQLite3 database stored on the filesystem
  
    **Requires**: ``privex-db`` (for both sync and async) and ``aiosqlite`` (only for async) libraries
    
    **Class:** :class:`.SqliteCache` (sync) // :class:`.AsyncSqliteCache` (async)


"""

DEFAULT_ASYNC_CACHE_ADAPTER = env('PRIVEX_ASYNC_CACHE_ADAPTER', DEFAULT_CACHE_ADAPTER)
"""
The default cache adapter for the AsyncIO caching system. :mod:`privex.helpers.cache` / :mod:`privex.helpers.cache.asyncx`

This defaults to :attr:`.DEFAULT_CACHE_ADAPTER` - since at the time of writing, all cache adapter alias names are also valid
for the AsyncIO cache system.

For example, when ``'redis'`` is specified to :attr:`.DEFAULT_ASYNC_CACHE_ADAPTER` , the :class:`.AsyncRedisCache` AsyncIO cache adapter
will be loaded for the async cache system, while when specified to the standard ``DEFAULT_CACHE_ADAPTER``, the
synchronous :class:`.RedisCache` adapter is loaded for use with the synchronous cache APIs.

Available adapters ( same as :attr:`.DEFAULT_CACHE_ADAPTER` ):

  * ``memory`` / ``mem`` / ``ram`` - Dependency-free memory cache inside of your application's memory.
  * ``redis`` - Caching via a local or remote Redis server
  * ``memcached`` / ``memcache`` / ``mcache`` - Caching via a local or remote Memcached server
  * ``sqlite3`` / ``sqlite`` / ``sqlitedb`` - Caching via an SQLite3 database stored on the filesystem


"""

########
# Redis Settings
########

REDIS_HOST = env('PRIVEX_REDIS_HOST', 'localhost')
"""Hostname / IP address where redis-server is running on"""
REDIS_PORT = _env_int('PRIVEX_REDIS_PORT', 6379)
"""Port number that Redis is running on at ``REDIS_HOST``"""
REDIS_DB = _env_int('PRIVEX_REDIS_DB', 0)
"""Redis database to use (number)"""

########
# Memcached Settings
########

MEMCACHED_HOST = env('PRIVEX_MEMCACHED_HOST', 'localhost')
"""Hostname / IP address where Memcached is running on"""
MEMCACHED_PORT = _env_int('PRIVEX_MEMCACHED_PORT', 11211)
"""Port number that Memcached is running on at ``MEMCACHED_HOST``"""

########################################
#                                      #
#       GeoIP Module Settings          #
#                                      #
#       privex.helpers.geoip           #
#                                      #
########################################

search_geoip = [
    '/usr/share/GeoIP',
    '/usr/lib/GeoIP',
    '/var/lib/GeoIP',
    '/usr/local/share/GeoIP',
    '/usr/local/var/GeoIP',
    '/var/GeoIP',
    join(getcwd(), 'GeoIP'),
    expanduser('~/GeoIP'),
    expanduser('~/.GeoIP'),
]


GEOIP_DIR = env('GEOIP_DIR', '/usr/share/GeoIP')
GEOASN_NAME = env('GEOASN_NAME', 'GeoLite2-ASN.mmdb')
GEOCITY_NAME = env('GEOCITY_NAME', 'GeoLite2-City.mmdb')
GEOCOUNTRY_NAME = env('GEOCOUNTRY_NAME', 'GeoLite2-Country.mmdb')

GEOASN_DETECTED, GEOCITY_DETECTED, GEOCOUNTRY_DETECTED = False, False, False

GEOCITY, GEOASN, GEOCOUNTRY = join(GEOIP_DIR, GEOCITY_NAME), join(GEOIP_DIR, GEOASN_NAME), join(GEOIP_DIR, GEOCOUNTRY_NAME)

TERMBIN_HOST, TERMBIN_PORT = 'termbin.com', 9999

CHECK_CONNECTIVITY: bool = _env_bool('CHECK_CONNECTIVITY', True)

HAS_WORKING_V4: Optional[bool] = None
"""
This is a storage variable - becomes either ``True`` or ``False`` after :func:`.check_v4` has been ran.

 * ``None`` - The connectivity checking function has never been ran - unsure where this IP version works or not.
 * ``True`` - This IP version appears to be fully functional - at least it was the last time the IP connectivity checking function was ran
 * ``False`` - This IP version appears to be broken - at least it was the last time the IP connectivity checking function was ran

"""
HAS_WORKING_V6: Optional[bool] = None
"""
This is a storage variable - becomes either ``True`` or ``False`` after :func:`.check_v6` has been ran.
 
 * ``None`` - The connectivity checking function has never been ran - unsure where this IP version works or not.
 * ``True`` - This IP version appears to be fully functional - at least it was the last time the IP connectivity checking function was ran
 * ``False`` - This IP version appears to be broken - at least it was the last time the IP connectivity checking function was ran

"""

SSL_VERIFY_CERT: bool = _env_bool('SSL_VERIFY_CERT', True)
SSL_VERIFY_HOSTNAME: bool = _env_bool('SSL_VERIFY_HOSTNAME', True)

DEFAULT_USER_AGENT = env('PRIVEX_USER_AGENT', "Python Privex Helpers ( https://github.com/Privex/python-helpers )")

# V4_CHECKED_AT: Optional[datetime] = None
# """
# This is a storage variable - used by :func:`.check_v4` to determine how long it's been since the host's IPv4 was tested.
# """
#
# V6_CHECKED_AT: Optional[datetime] = None
# """
# This is a storage variable - used by :func:`.check_v6` to determine how long it's been since the host's IPv6 was tested.
# """

NET_CHECK_TIMEOUT: int = _env_int('NET_CHECK_TIMEOUT', 3600)
"""
Number of seconds to cache the functional status of an IP version (caching applies to both positive and negative test results).
"""

NET_CHECK_HOST_COUNT: int = _env_int('NET_CHECK_HOST_COUNT', 3)
"""
Number of hosts in :attr:`.V4_TEST_HOSTS` / :attr:`.V6_TEST_HOSTS` that must be accessible - before that IP protocol
is considered functional.
"""

NET_CHECK_HOST_COUNT_TRY: int = _env_int('NET_CHECK_HOST_COUNT', 8)
"""
Maximum number of hosts in :attr:`.V4_TEST_HOSTS` / :attr:`.V6_TEST_HOSTS` that will be tested by :func:`.check_v4` / :func:`.check_v6`
"""

V4_TEST_HOSTS = [
    '185.130.44.10:80', '8.8.4.4:53', '1.1.1.1:53', '185.130.44.20:53', 'privex.io:80', 'files.privex.io:80',
    'google.com:80', 'www.microsoft.com:80', 'facebook.com:80', 'python.org:80'
]

V6_TEST_HOSTS = [
    '2a07:e00::333:53', '2001:4860:4860::8888:53', '2606:4700:4700::1111:53', '2a07:e00::abc:80',
    'privex.io:80', 'files.privex.io:80', 'google.com:80', 'facebook.com:80', 'bitbucket.org:80'
]

random.shuffle(V4_TEST_HOSTS)
random.shuffle(V6_TEST_HOSTS)

DEFAULT_SOCKET_TIMEOUT = 45

DEFAULT_READ_TIMEOUT = _env_int('DEFAULT_READ_TIMEOUT', 60)
DEFAULT_WRITE_TIMEOUT = _env_int('DEFAULT_WRITE_TIMEOUT', DEFAULT_READ_TIMEOUT)

DEFAULT_READ_TIMEOUT = None if DEFAULT_READ_TIMEOUT == 0 else DEFAULT_READ_TIMEOUT
DEFAULT_WRITE_TIMEOUT = None if DEFAULT_WRITE_TIMEOUT == 0 else DEFAULT_WRITE_TIMEOUT


def _bdir_plus_fname(f, sep='-', rem_ext=True):
    bpath, fname = path.split(f)
    _, bname = path.split(bpath)
    if rem_ext:
        fname, _ = path.splitext(fname)
    return f"{bname}{sep}{fname}"
    

SQLITE_APP_DB_NAME = env('SQLITE_APP_DB_NAME', _bdir_plus_fname(abspath(sys.argv[0])))
"""
This environment variable controls the default database file "name" portion for the SQLite3 database
wrappers in :mod:`privex.helpers.cache.post_deps` (used by the sync and async sqlite3 cache adapters)

If not set, it defaults to a combination of the **containing folder name** and the **base filename** of the currently
executing script - with the file extension (e.g. ``.py``) removed if it's present.

The folder name + base filename is retrieved from ``sys.argv[0]``. For example, if you ran an
application / script via ``./some/app/example.py``, then this would default to: ``app-example``

If you set this variable in your environment, for example, ``SQLITE_APP_DB_NAME=my_app`` would result in a default DB path
which looks like::

    /home/exampleuser/.privex_cache/my_app.sqlite3


"""

SQLITE_APP_DB_FOLDER = expanduser(env('SQLITE_APP_DB_FOLDER', '~/.privex_cache'))
"""
This environment variable controls the default base folder used to store database files for the SQLite3 database
wrappers in :mod:`privex.helpers.cache.post_deps` (used by the sync and async sqlite3 cache adapters).

The variable is only used if a user / application either doesn't specify a path to a DB file with the sync/async SQLite3
cache adapters, or they specify a relative path ( e.g. ``my_app/cache/cache_db.sqlite3`` ).

You may reference the current user's home directory (the user currently executing the python app/script that uses privex-helpers)
simply by using the character ``~``. For example, on a standard Linux system, if the executing user was ``john``, and
the ``SQLITE_APP_DB_FOLDER`` is set to ``~/.example``, then the folder would resolve to ``/home/john/.example``.

If not set, it defaults to ``~/.privex_cache``, a hidden folder directly within the current user's home directory.

This env var is used in conjunction with :attr:`.SQLITE_APP_DB_NAME` to produce a path to an SQLite3

"""



import asyncio
import socket
import ssl

from privex.helpers import settings
from privex.helpers.common import byteify, empty

from privex.helpers.net.util import generate_http_request, get_ssl_context, ip_is_v6, sock_ver

from privex.helpers.types import AnyNum, IP_OR_STR
from privex.helpers.net.dns import resolve_ip, resolve_ip_async
import logging

log = logging.getLogger(__name__)


def _ssl_context(kwargs, ssl_params=None):
    if not ssl_params:
        ssl_params = kwargs.pop(
            'ssl_params', dict(verify_cert=settings.SSL_VERIFY_CERT, check_hostname=settings.SSL_VERIFY_HOSTNAME)
        )
    return get_ssl_context(**ssl_params)


def _wrap_socket(s: socket.socket, kwargs: dict, host=None, wrap_params=None, ssl_params=None, ) -> ssl.SSLSocket:
    if not wrap_params:
        wrap_params = kwargs.pop('wrap_params', dict(
            server_hostname=kwargs.get('server_hostname', host),
            session=kwargs.get('session'),
            do_handshake_on_connect=kwargs.get('do_handshake_on_connect', True)
        ))
    ctx = _ssl_context(kwargs, ssl_params)
    return ctx.wrap_socket(s, **wrap_params)


def check_host(host: IP_OR_STR, port: AnyNum, version='any', throw=False, **kwargs) -> bool:
    """
    Test if the service on port ``port`` for host ``host`` is working. AsyncIO version: :func:`.check_host_async`

    Basic usage (services which send the client data immediately after connecting)::

        >>> check_host('hiveseed-se.privex.io', 2001)
        True
        >>> check_host('hiveseed-se.privex.io', 9991)
        False

    For some services, such as HTTP - it's necessary to transmit some data to the host before it will
    send a response. Using the ``send`` kwarg, you can transmit an arbitrary string/bytes upon connection.

    Sending data to ``host`` after connecting::

        >>> check_host('files.privex.io', 80, send=b"GET / HTTP/1.1\\n\\n")
        True


    :param str|IPv4Address|IPv6Address host: Hostname or IP to test
    :param int|str port: Port number on ``host`` to connect to
    :param str|int version: When connecting to a hostname, this can be set to ``'v4'``, ``'v6'`` or similar
                            to ensure the connection is via that IP version

    :param bool throw: (default: ``False``) When ``True``, will raise exceptions instead of returning ``False``
    :param kwargs: Additional configuration options (see below)

    :keyword int receive: (default: ``100``) Amount of bytes to attempt to receive from the server (``0`` to disable)
    :keyword bytes|str send: If ``send`` is specified, the data in ``send`` will be transmitted to the server before receiving.
    :keyword int stype: Socket type, e.g. :attr:`socket.SOCK_STREAM`

    :keyword float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                                If the global default timeout is ``None``, then falls back to ``5.0``

    :raises socket.timeout: When ``throw=True`` and a timeout occurs.
    :raises socket.gaierror: When ``throw=True`` and various errors occur
    :raises ConnectionRefusedError: When ``throw=True`` and the connection was refused
    :raises ConnectionResetError: When ``throw=True`` and the connection was reset

    :return bool success: ``True`` if successfully connected + sent/received data. Otherwise ``False``.
    """
    kwargs = dict(kwargs)
    receive, stype = int(kwargs.get('receive', 100)), kwargs.get('stype', socket.SOCK_STREAM)
    timeout, send, use_ssl = kwargs.get('timeout', 'n/a'), kwargs.get('send'), kwargs.get('ssl', kwargs.get('use_ssl'))
    http_test, hostname = kwargs.get('http_test', False), kwargs.get('hostname', host)

    # ssl_params = kwargs.get('ssl_params', dict(verify_cert=False, check_hostname=False))
    if timeout == 'n/a':
        t = socket.getdefaulttimeout()
        timeout = settings.DEFAULT_SOCKET_TIMEOUT if not t else t
    if http_test:
        send = generate_http_request(url=kwargs.get('url', '/'), host=hostname)
    try:
        s_ver = socket.AF_INET
        ip = resolve_ip(host, version)
        
        if ip_is_v6(ip): s_ver = socket.AF_INET6
        
        if port == 443 and use_ssl is None:
            log.warning("check_host: automatically setting use_ssl=True as port is 443 and use_ssl was not specified.")
            use_ssl = True
        with socket.socket(s_ver, stype) as s:
            if use_ssl: s = _wrap_socket(s, kwargs, host)
            if timeout: s.settimeout(float(timeout))
            
            s.connect((ip, int(port)))
            if not empty(send):
                s.sendall(byteify(send))
            if receive > 0:
                s.recv(int(receive))
            if use_ssl:
                s.close()
        return True
    except (socket.timeout, TimeoutError, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        if throw:
            raise e
    return False


async def check_host_async(host: IP_OR_STR, port: AnyNum, version='any', throw=False, **kwargs) -> bool:
    """
    AsyncIO version of :func:`.check_host`. Test if the service on port ``port`` for host ``host`` is working.

    Basic usage (services which send the client data immediately after connecting)::

        >>> await check_host_async('hiveseed-se.privex.io', 2001)
        True
        >>> await check_host_async('hiveseed-se.privex.io', 9991)
        False

    For some services, such as HTTP - it's necessary to transmit some data to the host before it will
    send a response. Using the ``send`` kwarg, you can transmit an arbitrary string/bytes upon connection.

    Sending data to ``host`` after connecting::

        >>> await check_host_async('files.privex.io', 80, send=b"GET / HTTP/1.1\\n\\n")
        True


    :param str|IPv4Address|IPv6Address host: Hostname or IP to test
    :param int|str port: Port number on ``host`` to connect to
    :param str|int version: When connecting to a hostname, this can be set to ``'v4'``, ``'v6'`` or similar
                            to ensure the connection is via that IP version

    :param bool throw: (default: ``False``) When ``True``, will raise exceptions instead of returning ``False``
    :param kwargs: Additional configuration options (see below)

    :keyword int receive: (default: ``100``) Amount of bytes to attempt to receive from the server (``0`` to disable)
    :keyword bytes|str send: If ``send`` is specified, the data in ``send`` will be transmitted to the server before receiving.
    :keyword int stype: Socket type, e.g. :attr:`socket.SOCK_STREAM`

    :keyword float|int timeout: Socket timeout. If not passed, uses the default from :func:`socket.getdefaulttimeout`.
                                If the global default timeout is ``None``, then falls back to ``5.0``

    :raises socket.timeout: When ``throw=True`` and a timeout occurs.
    :raises socket.gaierror: When ``throw=True`` and various errors occur
    :raises ConnectionRefusedError: When ``throw=True`` and the connection was refused
    :raises ConnectionResetError: When ``throw=True`` and the connection was reset

    :return bool success: ``True`` if successfully connected + sent/received data. Otherwise ``False``.
    """
    kwargs = dict(kwargs)
    receive, stype = int(kwargs.get('receive', 100)), kwargs.get('stype', socket.SOCK_STREAM)
    timeout, send, use_ssl = kwargs.get('timeout', 'n/a'), kwargs.get('send'), kwargs.get('ssl', kwargs.get('use_ssl'))
    http_test, hostname = kwargs.get('http_test', False), kwargs.get('hostname', host)
    
    if timeout == 'n/a':
        t = socket.getdefaulttimeout()
        timeout = 10.0 if not t else t
    
    loop = asyncio.get_event_loop()
    if http_test:
        send = generate_http_request(url=kwargs.get('url', '/'), host=hostname)
    if port == 443 and use_ssl is None:
        log.warning("check_host_async: automatically setting use_ssl=True as port is 443 and use_ssl was not specified.")
        use_ssl = True
    try:
        if sock_ver(version) is None:
            s_ver = socket.AF_INET
            host = await resolve_ip_async(host, version)
            if ip_is_v6(host): s_ver = socket.AF_INET6
        else:
            s_ver = sock_ver(version)
        
        with socket.socket(s_ver, stype) as s:
            if use_ssl: s = _wrap_socket(s, kwargs, host)
            if timeout:
                s.settimeout(float(timeout))
                await asyncio.wait_for(loop.sock_connect(s, (host, int(port))), timeout)
            else:
                await loop.sock_connect(s, (host, int(port)))
            
            if not empty(send):
                await loop.sock_sendall(s, byteify(send))
            if receive > 0:
                await loop.sock_recv(s, int(receive))
        return True
    except (socket.timeout, TimeoutError, ConnectionRefusedError, ConnectionResetError, socket.gaierror) as e:
        if throw:
            raise e
    return False



import warnings

from privex.helpers.decorators import mock_decorator
from privex.helpers.collections import Mocker, dataclasses_mock

module = Mocker.make_mock_class('module')


def mkclass(name: str = 'module', instance: bool = True, **kwargs):
    return Mocker.make_mock_class(name, instance=instance, **kwargs)


def mkmodule(mod_name: str, attributes: dict = None, modules: dict = None, **kwargs):
    return Mocker.make_mock_module(mod_name, attributes, modules, **kwargs)


dataclasses = dataclasses_mock
dataclass, field = dataclasses.dataclass, dataclasses.field

pytest = mkmodule(
    'pytest',
    dict(
        skip=lambda msg, allow_module_level=True: warnings.warn(msg),
        mark=Mocker.make_mock_class(
            '_pytest.mark.structures.MarkGenerator',
            attributes=dict(skip=mock_decorator, skipif=mock_decorator())
        )
    )
)

attr = Mocker(
    attributes=dict(
        s=mock_decorator,
        asdict=lambda obj, dict_factory=dict: dict_factory(obj),
        astuple=lambda obj, tuple_factory=tuple: tuple_factory(obj),
        validate=lambda obj: False
    )
)





"""
This module handles connection objects for databases, APIs etc. by exposing functions which initialise and
store class instances for re-use.

It's primarily intended to be used to enable database, caching and third-party API connectivity for the helpers in this
package, however, you're free to use the functions / classes / attributes exposed in this module for your own apps.

Classes are generally initialised using the settings from :py:mod:`.settings` - see the docs for that module to learn
how to override the settings if the defaults don't work for you.


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import logging
import threading
from os import path
from typing import Any, Union, Optional, Generator, Tuple, Dict

from privex.helpers.collections import DictObject

from privex.helpers import settings
from privex.helpers.types import T
from privex.helpers.common import empty_if, empty
from privex.helpers.exceptions import GeoIPDatabaseNotFound

log = logging.getLogger(__name__)

__all__ = [
    'HAS_REDIS', 'HAS_ASYNC_REDIS', 'HAS_ASYNC_MEMCACHED', 'HAS_DNSPYTHON', 'HAS_CRYPTO', 'HAS_SETUPPY_BUMP',
    'HAS_SETUPPY_COMMANDS', 'HAS_SETUPPY_COMMON', 'HAS_GEOIP', 'HAS_MEMCACHED', 'HAS_PRIVEX_DB', 'clean_threadstore'
]

HAS_REDIS = False
"""If the ``redis`` module was imported successfully, this will change to True."""

HAS_ASYNC_REDIS = False
"""If the ``aioredis`` module was imported successfully, this will change to True."""

HAS_ASYNC_MEMCACHED = False
HAS_DNSPYTHON = False
"""If the ``dns.resolver`` module was imported successfully, this will change to True."""

HAS_CRYPTO = False
"""If :py:mod:`privex.helpers.crypto` was imported successfully, this will change to True"""

HAS_SETUPPY_COMMON = False
"""If :py:mod:`privex.helpers.setuppy.common` was imported successfully, this will change to True"""

HAS_SETUPPY_BUMP = False
"""If :py:mod:`privex.helpers.setuppy.bump` was imported successfully, this will change to True"""

HAS_SETUPPY_COMMANDS = False
"""If :py:mod:`privex.helpers.setuppy.commands` was imported successfully, this will change to True"""

HAS_GEOIP = False
"""If :py:mod:`privex.helpers.geoip` was imported successfully, this will change to True"""

HAS_MEMCACHED = False

HAS_PRIVEX_DB = None
"""If the ``privex.db`` module was imported successfully, this will change to True."""

__STORE = dict(threads={})
"""This ``dict`` is used to store initialised classes for connections to databases, APIs etc."""


def _get_threadstore(name=None, fallback=None, thread_id=None) -> Any:
    thread_id = empty_if(thread_id, threading.get_ident())
    if thread_id not in __STORE['threads']:
        __STORE['threads'][thread_id] = {}
    thread_store: dict = __STORE['threads'][thread_id]
    if name is None:
        return thread_store
    
    return thread_store.get(name, fallback)


def _get_all_threadstore(name=None) -> Generator[Tuple[Union[int, str], Union[Dict[str, Any], Any]], None, None]:
    """
    
    Get the ``'redis'`` key from every thread in the thread store::
    
        >>> for t_id, inst in _get_all_threadstore('redis'):
        ...     inst.close()
        ...
        >>> clean_threadstore(name='redis')
    
    Get the thread store for every single thread::
        
        >>> for t_id, ts in _get_all_threadstore():
        ...     if 'redis' in ts: print(t_id, 'has redis instance!')
        ...     if 'aiomemcached' in ts: print(t_id, 'has asyncio memcached instance!')
    
    
    :param str name: Yield only this key from each thread store (if it exists)
    :return Generator store_gen: A generator of tuples containing either ``(thread_id, thread_store: dict)``, or ``(t_id, value)``
                                 depending on whether ``name`` is empty or not.
    """
    for t_id, ts in __STORE['threads'].items():
        if empty(name):
            yield t_id, ts
        elif name in ts:
            yield t_id, ts[name]


def _set_threadstore(name, obj: T, thread_id=None) -> T:
    thread_store = _get_threadstore(thread_id=thread_id)
    thread_store[name] = obj
    return obj


def clean_threadstore(thread_id=None, name=None, clean_all: bool = False) -> bool:
    """
    Remove the per-thread instance storage in :attr:`.__STORE`, usually called when a thread is exiting.

    Can also be used to clear a certain key in all thread stores, or completely clear every key from every thread store.

    Example::

        >>> def some_thread():
        ...     r = get_redis()
        ...     print('doing something')
        ...     print('cleaning up...')
        ...     clean_threadstore()       # With no arguments, it cleans the thread store for the thread that called it.
        >>> t = threading.Thread(target=some_thread)
        >>> t.start()
        >>> t.join()

    Usage outside of a thread::

        >>> t = threading.Thread(target=some_thread)
        >>> t.start()
        >>> thread_id = t.ident                      # Get the thread ID for the started thread
        >>> t.join()                                 # Wait for the thread to finish
        >>> if thread_id is not None:                # Make sure the thread ID isn't None
        ...     clean_threadstore(thread_id)         # Cleanup any leftover instances, if there are any.
        ...

    Removing an individual item from thread store::
        
        >>> def some_thread():
        ...     r = get_redis()
        ...     print('doing something')
        ...     print('cleaning up...')
        ...     clean_threadstore(name='redis')   # Delete only the key 'redis' from the thread store
    
    Removing an individual item from the thread store for **ALL thread ID's**::
    
        >>> # Remove the redis instance from every thread store
        >>> clean_threadstore(name='redis', clean_all=True)
    
    Clearing the entire thread store for every thread ID::
    
        >>> clean_threadstore(clean_all=True)
    
    
    :param thread_id: The ID of the thread (usually from :func:`threading.get_ident`) to clean the storage for.
                      If left as None, will use the ID returned by :func:`threading.get_ident`.
    
    :param name:      If specified, then only the key ``name`` will be deleted from the thread store, instead of the entire thread store.
    :param bool clean_all: (default: ``False``) If ``True`` - when ``name`` is non-empty - that key will be removed from every
                           thread ID's thread store - while if ``name`` is empty, every single thread ID's thread store is cleared.
    """
    thread_id = empty_if(thread_id, threading.get_ident())

    if clean_all:
        for t_id, ts in __STORE['threads'].items():
            ts: dict
            if empty(name):
                log.debug("[clean_threadstore] (clean_all True) Cleaning entire thread store for thread ID '%s'", t_id)
                for n in ts.keys():
                    log.debug("[clean_threadstore] Cleaning '%s' key for thread ID '%s'", name, t_id)
                    del ts[n]
                continue
            log.debug("[clean_threadstore] (clean_all True) Cleaning '%s' key for thread ID '%s'", name, t_id)
            if name in ts:
                log.debug("[clean_threadstore] Found %s in thread ID %s - deleting...", name, ts)
                del ts[name]
        return True
    
    if thread_id in __STORE['threads']:
        ts = __STORE['threads'][thread_id]
        if name is None:
            del __STORE['threads'][thread_id]
            return True
        elif name in ts:
            del ts[name]
            return True
    return False


try:
    import redis


    def connect_redis(*args, **rd_config):
        from privex.helpers.common import extract_settings
        # return redis.Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT, db=settings.REDIS_DB)
        return redis.Redis(*args, **extract_settings('REDIS_', settings, merge_conf=dict(rd_config)))
    
    def get_redis(new_connection=False, thread_id=None, **rd_config) -> redis.Redis:
        """Get a Redis connection object. Create one if it doesn't exist."""
        from privex.helpers.common import empty_if, extract_settings

        if new_connection:
            return connect_redis(**rd_config)
        # if thread_id not in __STORE['threads']: __STORE['threads'][thread_id] = {}

        # thread_store = __STORE['threads'][thread_id]
        rd = _get_threadstore('redis', thread_id=thread_id)
        
        if rd is None:
            rd = connect_redis(**rd_config)
            _set_threadstore('redis', rd, thread_id=thread_id)
        return rd
        # if 'redis' in thread_store and not new_connection: return thread_store['redis']
        # if rd is not None and not new_connection:
        #
        # # if 'redis' not in thread_store:
        #     if new_connection:
        #         return rd
        #     thread_store['redis'] = rd
        # return thread_store['redis']
    
    def close_redis(thread_id=None, close_all=False) -> bool:
        """
        Close the global Redis connection and delete the instance.
        
        :param thread_id: Close and delete the Redis instance for this thread ID, instead of the detected current thread
        :param bool close_all: Close all AsyncIO Redis instances for every thread ID, then purge their thread store keys.
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Redis instance.
        """
        if close_all:
            for t_id, rd in _get_all_threadstore('redis'):
                if rd is None:
                    continue
                log.debug("Closing redis for thread ID %s", t_id)
                try:
                    rd.close()
                except Exception:
                    log.exception("Exception while closing redis for thread ID %s", t_id)
            return clean_threadstore(name='redis', clean_all=True)
    
        rd: Union[str, redis.Redis] = _get_threadstore('redis', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            rd.close()
            clean_threadstore(thread_id=thread_id, name='redis')
            return True
        return False

    def reset_redis(thread_id=None) -> redis.Redis:
        """Close the global Redis connection, delete the instance, then re-instantiate it"""
        close_redis(thread_id=thread_id)
        return _set_threadstore('redis', get_redis(True, thread_id), thread_id=thread_id)
    
    def configure_redis(host=settings.REDIS_HOST, port: int = settings.REDIS_PORT, db: int = settings.REDIS_DB, **kwargs):
        """Update global Redis settings and re-instantiate the global Redis instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.REDIS_DB = db
        settings.REDIS_PORT = port
        settings.REDIS_HOST = host
        return reset_redis(thread_id=thread_id)


    __all__ += ['get_redis', 'reset_redis', 'configure_redis', 'close_redis', 'connect_redis']
    HAS_REDIS = True

except ImportError:
    log.debug('privex.helpers __init__ failed to import "redis", Redis dependent helpers will be disabled.')
except Exception as e:
    log.debug('privex.helpers __init__ failed to import "redis", (unknown exception), '
              'Redis dependent helpers will be disabled. Exception: %s %s', type(e), str(e))

try:
    import aioredis
    HAS_ASYNC_REDIS = True


    async def connect_redis_async(**rd_config) -> aioredis.Redis:
        from privex.helpers.common import extract_settings
        cf = extract_settings('REDIS_', settings, merge_conf=dict(rd_config))
        _addr = f"redis://{cf.pop('host', 'localhost')}:{cf.pop('port', 6379)}"
        return await aioredis.create_redis_pool(
            address=_addr, **cf
        )


    async def get_redis_async(new_connection=False, thread_id=None, **rd_config) -> aioredis.connection:
        """
        Get an Async Redis connection object. Create one if it doesn't exist.
        
        Usage::
            
            >>> redis_conn = await get_redis_async()
            >>> redis = await redis_conn
            >>> await redis.set('some_key', 'example')
            >>> await redis.get('some_key')
            'example'
            
        
        """
        from privex.helpers.common import empty_if, extract_settings
    
        if new_connection: return await connect_redis_async(**rd_config)
        rd: Optional[aioredis.Redis] = _get_threadstore('aioredis', thread_id=thread_id)
    
        if rd is None:
            rd: aioredis.Redis = _set_threadstore('aioredis', await connect_redis_async(**rd_config), thread_id=thread_id)
        return rd

    async def close_redis_async(thread_id=None, close_all=False) -> bool:
        """
        Close the global Async Redis connection and delete the instance.

        :param thread_id: Close and delete the Redis instance for this thread ID, instead of the detected current thread
        :param bool close_all: Close all AsyncIO Redis instances for every thread ID, then purge their thread store keys.
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Redis instance.
        """
        rd: Optional[Union[aioredis.connection, aioredis.Redis]]
        if close_all:
            for t_id, rd in _get_all_threadstore('aioredis'):
                if rd is None:
                    continue
                log.debug("Closing aioredis for thread ID %s", t_id)
                try:
                    rd.close()
                    await rd.wait_closed()
                except RuntimeError as err:
                    log.warning("[ignored] RuntimeError while closing aioredis (thread %s): %s - %s", thread_id, type(err), str(err))
                except Exception:
                    log.exception("Exception while closing aioredis for thread ID %s", t_id)
            return clean_threadstore(name='aioredis', clean_all=True)
        
        rd = _get_threadstore('aioredis', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            try:
                rd.close()
                await rd.wait_closed()
            except RuntimeError as err:
                log.warning("[ignored] RuntimeError while closing aioredis (thread %s): %s - %s", thread_id, type(err), str(err))
            clean_threadstore(thread_id=thread_id, name='aioredis')
            return True
        return False


    async def reset_redis_async(thread_id=None) -> aioredis.connection:
        """Close the global Async Redis connection, delete the instance, then re-instantiate it"""
        await close_redis_async(thread_id=thread_id)
        return await _set_threadstore('aioredis', get_redis_async(True, thread_id), thread_id=thread_id)


    async def configure_redis_async(host=settings.REDIS_HOST, port: int = settings.REDIS_PORT, db: int = settings.REDIS_DB, **kwargs):
        """Update global Redis settings and re-instantiate the global Async Redis instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.REDIS_DB = db
        settings.REDIS_PORT = port
        settings.REDIS_HOST = host
        return await reset_redis_async(thread_id=thread_id)


    __all__ += ['get_redis_async', 'reset_redis_async', 'configure_redis_async', 'close_redis_async', 'connect_redis_async']


except ImportError:
    log.debug('privex.helpers __init__ failed to import "aioredis", Async Redis dependent helpers will be disabled.')
except Exception as e:
    log.debug('privex.helpers __init__ failed to import "aioredis", (unknown exception), '
              'Async Redis dependent helpers will be disabled. Exception: %s %s', type(e), str(e))

try:
    import aiomcache
    HAS_ASYNC_MEMCACHED = True

    async def connect_memcached_async(**rd_config) -> aiomcache.Client:
        from privex.helpers.common import extract_settings
        cf = extract_settings('MEMCACHED_', settings, merge_conf=dict(rd_config))
        # if 'loop' not in cf:
        #     cf['loop'] = asyncio.get_event_loop()
        return aiomcache.Client(**cf)


    async def get_memcached_async(new_connection=False, thread_id=None, **rd_config) -> aiomcache.Client:
        """Get an Async Memcached connection object. Create one if it doesn't exist."""
        from privex.helpers.common import empty_if, extract_settings
    
        if new_connection: return await connect_memcached_async(**rd_config)
        rd: Optional[aiomcache.Client] = _get_threadstore('aiomemcached', thread_id=thread_id)
    
        if rd is None:
            rd: aiomcache.Client = _set_threadstore('aiomemcached', await connect_memcached_async(**rd_config), thread_id=thread_id)
        return rd


    async def close_memcached_async(thread_id=None, close_all=False) -> bool:
        """
        Close the global Async Memcached connection and delete the instance.

        :param thread_id: Close and delete the Memcached instance for this thread ID, instead of the detected current thread
        :param bool close_all: Close all AsyncIO Memcached instances for every thread ID, then purge their thread store keys.
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Memcached instance.
        """
        if close_all:
            for t_id, rd in _get_all_threadstore('aiomemcached'):
                if rd is None:
                    continue
                log.debug("Closing aiomemcached for thread ID %s", t_id)
                try:
                    await rd.close()
                except Exception:
                    log.exception("Exception while closing aiomemcached for thread ID %s", t_id)
            return clean_threadstore(name='aiomemcached', clean_all=True)
        
        rd: Union[str, aiomcache.Client] = _get_threadstore('aiomemcached', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            await rd.close()
            clean_threadstore(thread_id=thread_id, name='aiomemcached')
            return True
        return False


    async def reset_memcached_async(thread_id=None) -> aiomcache.Client:
        """Close the global Async Memcached connection, delete the instance, then re-instantiate it"""
        await close_memcached_async(thread_id=thread_id)
        return await _set_threadstore('aiomemcached', get_memcached_async(True, thread_id), thread_id=thread_id)


    async def configure_memcached_async(host=settings.MEMCACHED_HOST, port: int = settings.MEMCACHED_PORT, **kwargs):
        """Update global Memcached settings and re-instantiate the global Async Redis instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.MEMCACHED_PORT = port
        settings.MEMCACHED_HOST = host
        return await reset_memcached_async(thread_id=thread_id)


    __all__ += ['get_memcached_async', 'reset_memcached_async', 'configure_memcached_async', 'close_memcached_async', 'connect_memcached_async']

except ImportError:
    log.debug('%s failed to import "aiomcache", Async Memcached dependent helpers will be disabled.', __name__)
except Exception as e:
    log.debug('%s failed to import "aiomcache", (unknown exception), Async Memcached dependent helpers '
              'will be disabled. Exception: %s %s', __name__, type(e), str(e))

try:
    import pylibmc
    
    HAS_MEMCACHED = True
    
    
    def connect_memcached(**rd_config) -> pylibmc.Client:
        from privex.helpers.common import extract_settings
        cf = extract_settings('MEMCACHED_', settings, merge_conf=dict(rd_config))
        # If MEMCACHED_HOST isn't a list/tuple/set, then we need to make it into one, since pylibmc expects a LIST of servers.
        # It also doesn't have an argument for a port, so we combine that with the host.
        host, port = cf.pop('host', 'localhost'), cf.pop('port', 11211)
        if isinstance(host, (list, tuple, set)):
            fhost = host
        else:
            fhost = [f"{host}:{port}" if port != 11211 else host]
        return pylibmc.Client(fhost,  **cf)
    
    
    def get_memcached(new_connection=False, thread_id=None, **rd_config) -> pylibmc.Client:
        """Get a synchronous Memcached connection object. Create one if it doesn't exist."""
        from privex.helpers.common import empty_if, extract_settings
        
        if new_connection: return connect_memcached(**rd_config)
        rd: Optional[pylibmc.Client] = _get_threadstore('memcached', thread_id=thread_id)
        
        if rd is None:
            rd: pylibmc.Client = _set_threadstore('memcached', connect_memcached(**rd_config), thread_id=thread_id)
        return rd
    
    
    def close_memcached(thread_id=None, close_all=False) -> bool:
        """
        Close the global synchronous Memcached connection and delete the instance.

        :param thread_id: Close and delete the Memcached instance for this thread ID, instead of the detected current thread
        :param bool close_all: Close all synchronous Memcached instances for every thread ID, then purge their thread store keys.
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Memcached instance.
        """
        if close_all:
            for t_id, rd in _get_all_threadstore('memcached'):
                if rd is None:
                    continue
                log.debug("Closing memcached for thread ID %s", t_id)
                rd: pylibmc.Client
                try:
                    rd.disconnect_all()
                except Exception:
                    log.exception("Exception while closing memcached for thread ID %s", t_id)
            return clean_threadstore(name='memcached', clean_all=True)
        
        rd: Union[str, pylibmc.Client] = _get_threadstore('memcached', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            rd.disconnect_all()
            clean_threadstore(thread_id=thread_id, name='memcached')
            return True
        return False
    
    
    def reset_memcached(thread_id=None) -> pylibmc.Client:
        """Close the global synchronous Memcached connection, delete the instance, then re-instantiate it"""
        close_memcached(thread_id=thread_id)
        return _set_threadstore('memcached', get_memcached(True, thread_id), thread_id=thread_id)
    
    
    def configure_memcached(host=settings.MEMCACHED_HOST, port: int = settings.MEMCACHED_PORT, **kwargs) -> pylibmc.Client:
        """Update global Memcached settings and re-instantiate the global Memcached instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.MEMCACHED_PORT = port
        settings.MEMCACHED_HOST = host
        return reset_memcached(thread_id=thread_id)
    
    
    __all__ += ['get_memcached', 'reset_memcached', 'configure_memcached', 'close_memcached',
                'connect_memcached']

except ImportError:
    log.debug('%s failed to import "pylibmc", Synchronous Memcached dependent helpers will be disabled.', __name__)
except Exception as e:
    log.debug('%s failed to import "pylibmc", (unknown exception), Synchronous Memcached dependent helpers '
              'will be disabled. Exception: %s %s', __name__, type(e), str(e))

try:
    import geoip2
    import geoip2.database

    def connect_geoip(*args, **geo_config) -> geoip2.database.Reader:
        return geoip2.database.Reader(*args, **geo_config)

    def get_geodbs() -> DictObject:
        return DictObject({
            'city':    DictObject(
                name=settings.GEOCITY_NAME, path=settings.GEOCITY, detected=settings.GEOCITY_DETECTED
            ),
            'country': DictObject(
                name=settings.GEOCOUNTRY_NAME, path=settings.GEOCOUNTRY, detected=settings.GEOCOUNTRY_DETECTED
            ),
            'asn': DictObject(
                name=settings.GEOASN_NAME, path=settings.GEOASN, detected=settings.GEOASN_DETECTED
            ),
        })

    def _find_geoip(geo_type: str) -> Optional[str]:
        _geo_dbs = get_geodbs()

        if geo_type not in _geo_dbs:
            raise AttributeError("_find_geoip: geo_type must be 'city', 'asn' or 'country'")
        log.debug("Locating GeoIP %s - _geo_dbs = %s", geo_type.capitalize(), _geo_dbs)
        for p in settings.search_geoip:
            g_p = path.join(p, _geo_dbs[geo_type].name)
            log.debug("Checking if %s exists...", g_p)
            if path.exists(g_p):
                log.debug("Found GeoIP %s (%s) in path: %s", geo_type, _geo_dbs[geo_type], g_p)
                return g_p
        log.warning("Failed to find GeoIP %s in search paths %s", geo_type.capitalize(), settings.search_geoip)
        return None
        

    def get_geoip_db(geo_type: str) -> str:
        """
        Return the full path to the GeoIP2 database for ``geo_type``.
        
        If we haven't yet scanned the search paths for the database, then :func:`._find_geoip` will be called
        to try and locate the database file.
        
        If the database is found, the ``_DETECTED`` boolean setting will be changed to ``True`` so we know that the
        path contained in the :func:`.get_geodbs` result is valid in the future, avoiding unnecessary searches.
        
        If the database can't be found anywhere within the search paths, :class:`.GeoIPDatabaseNotFound` will be raised.
        
        :param str geo_type: The GeoIP database type: either 'city', 'asn' or 'country'
        :raises GeoIPDatabaseNotFound: If the database for ``geo_type`` could not be found.
        :return str path: The full path to the detected GeoIP database
        """
        _geo_dbs = get_geodbs()
    
        if geo_type not in _geo_dbs:
            raise AttributeError("get_geoip_db: geo_type must be 'city', 'asn' or 'country'")
        gdb = _geo_dbs[geo_type]
        if gdb.detected:
            return gdb.path
        
        gp = _find_geoip(geo_type)
        if not gp:
            log.error("Failed to locate GeoIP ASN + City in alternative search folders.")
            log.error(f" [!!!] ERROR - Missing GeoIP files. The following files do not exist:\n")
            log.error(f"\t{gdb.path}")
            log.error(f"\nCurrent environment settings for GeoIP location:\n")
    
            log.error(f"GEOIP_DIR={settings.GEOIP_DIR}")
            log.error(f"GEOASN_NAME={settings.GEOASN_NAME}")
            log.error(f"GEOCITY_NAME={settings.GEOCITY_NAME}\n")
            log.error(f"GEOCOUNTRY_NAME={settings.GEOCOUNTRY_NAME}\n")
    
            log.error("Please download the GeoLite2 City and ASN GeoIP databases (Maxmind Binary Database format) "
                      "for FREE from the following URL:\n")
            log.error("\thttps://dev.maxmind.com/geoip/geoip2/geolite2/\n")
            log.error("It's recommended to place the .mmdb files in the default folder '/usr/share/GeoIP/' - "
                      "as most applications which use Maxmind GeoIP2 will use that folder by default.\n")
            log.error("You can alternatively place the .mmdb files into any of these folders:\n")
            for p in settings.search_geoip:
                log.error("\t - %s", p)
            log.error("")
            log.error("For most functionality, you'll need both GeoIP2 City + GeoIP2 ASN. The GeoIP2 City database generally provides "
                      "everything in the Country database, plus City data - making the Country database generally redundant.\n")
            raise GeoIPDatabaseNotFound(f"Failed to locate GeoIP {geo_type.capitalize()} ({gdb.name})")
        
        if geo_type == 'city':
            settings.GEOCITY, settings.GEOCITY_DETECTED = gp, True
        if geo_type == 'country':
            settings.GEOCOUNTRY, settings.GEOCOUNTRY_DETECTED = gp, True
        if geo_type == 'asn':
            settings.GEOASN, settings.GEOASN_DETECTED = gp, True
        
        return gp
        

    def get_geoip(geo_type: str, new_connection=False, thread_id=None, **geo_config) -> geoip2.database.Reader:
        """Get a GeoIP Reader object. Create one if it doesn't exist."""
        from privex.helpers.common import empty_if, extract_settings
        
        gdb = get_geoip_db(geo_type)
        _geo_dbs = get_geodbs()
        
        if geo_type not in _geo_dbs:
            raise AttributeError("get_geoip: geo_type must be 'city', 'asn' or 'country'")
        
        if new_connection:
            return connect_geoip(gdb, **geo_config)

        tsname = f'geoip_{geo_type}'
        geo = _get_threadstore(tsname, thread_id=thread_id)
    
        if geo is None:
            geo = connect_geoip(gdb, **geo_config)
            _set_threadstore(tsname, geo, thread_id=thread_id)
        return geo


    def close_geoip(geo_type: str, thread_id=None, close_all=False) -> bool:
        """
        Close the global GeoIP connection and delete the instance.
        
        :param str geo_type: The GeoIP database type: either 'city', 'asn' or 'country'
        :param thread_id: Close and delete the Redis instance for this thread ID, instead of the detected current thread
        :param bool close_all: Close all GeoIP ``geo_type`` instances for every thread ID, then purge their thread store keys.
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Redis instance.
        """
        tsname = f'geoip_{geo_type}'

        if close_all:
            for t_id, geo in _get_all_threadstore(tsname):
                if geo is None:
                    continue
                log.debug("Closing %s for thread ID %s", tsname, t_id)
                try:
                    geo.close()
                except Exception:
                    log.exception("Exception while closing %s for thread ID %s", tsname, t_id)
            return clean_threadstore(name=tsname, clean_all=True)
        
        geo: Union[str, geoip2.database.Reader] = _get_threadstore(tsname, 'NOT_FOUND', thread_id=thread_id)
        if geo != 'NOT_FOUND':
            geo.close()
            clean_threadstore(thread_id=thread_id, name=tsname)
            return True
        return False


    def reset_geoip(geo_type: str, thread_id=None) -> geoip2.database.Reader:
        """Close the global GeoIP connection, delete the instance, then re-instantiate it"""
        close_geoip(geo_type, thread_id=thread_id)
        return get_geoip(geo_type, thread_id=thread_id)


    __all__ += ['get_geoip', 'get_geoip_db', 'get_geodbs', 'reset_geoip', 'close_geoip', 'connect_geoip']
    HAS_GEOIP = True
except ImportError:
    log.debug('%s failed to import "geoip2", GeoIP2 dependent helpers will be disabled.', __name__)





